/* ------------------------------------------------------------------------
System       : cc65 cross-development environment
Platform     : Apple IIgs or Apple IIe with VOC or CB

Bill Buckels
July 2014

Licence      : You may use this program for whatever you wish as long
			   as you agree that Bill Buckels has no warranty or
			   liability obligations whatsoever from said use.
------------------------------------------------------------------------ */

/*

1. reads a 24 bit BMP in resolutions from 2 x 1 to 320 x 200
2. reads palette 0 from a 320 x 200 SHR Raw PIC background image
   if no background image, a built-in palette is used.
   the built-in palettes are 16-color DHGR palettes in the Lo-Res color order
3. writes an SHR raw sprite file of image data with a 16 color palette
   ProDOS File Type $C1 Auxiliary Type $CC65

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <math.h>

/* ***************************************************************** */
/* ========================== defines ============================== */
/* ***************************************************************** */

#define SUCCESS 0
#define INVALID -1

/* constants for the biCompression field in a BMP header */
#define BI_RGB      0L
#define BI_RLE8     1L
#define BI_RLE4     2L

/* ***************************************************************** */
/* ========================== typedefs ============================= */
/* ***************************************************************** */

typedef unsigned char uchar;
typedef unsigned short ushort;
typedef unsigned long ulong;
typedef short sshort;

/* Bitmap Header structures */
#ifdef MINGW
typedef struct __attribute__((__packed__)) tagBITMAPINFOHEADER
#else
typedef struct tagBITMAPINFOHEADER
#endif
{
    ulong   biSize;
    ulong   biWidth;
    ulong   biHeight;
    ushort  biPlanes;
    ushort  biBitCount;
    ulong   biCompression;
    ulong   biSizeImage;
    ulong   biXPelsPerMeter;
    ulong   biYPelsPerMeter;
    ulong   biClrUsed;
    ulong   biClrImportant;
} BITMAPINFOHEADER;

#ifdef MINGW
typedef struct __attribute__((__packed__)) tagBITMAPFILEHEADER
#else
typedef struct tagBITMAPFILEHEADER
#endif
{
    uchar   bfType[2];
    ulong   bfSize;
    ushort  bfReserved1;
    ushort  bfReserved2;
    ulong   bfOffBits;
} BITMAPFILEHEADER;


/* ***************************************************************** */
/* ========================== globals ============================== */
/* ***************************************************************** */

/* static structures for processing */
BITMAPFILEHEADER bfi;
BITMAPINFOHEADER bmi;

char bmpfile[256], spritefile[256], picfile[256], fname[256];
int writeheader = 1, writepalette = 1, backgroundcolor = -1, quietmode = 1;

/* 4 - built-in palette options */
/* based on DHGR, DLGR, LGR colors */
/* LGR color order */

unsigned char kegs32colors[16][3] = {
	0,0,0,		 /* black    */
	221,0,51,	 /* red      */
	0,0,153,	 /* dk blue  */
	221,0,221,	 /* purple   */
	0,119,0,	 /* dk green */
	85,85,85,	 /* gray     */
	34,34,255,	 /* med blue */
	102,170,255, /* lt blue  */
	136,85,34,	 /* brown    */
	255,102,0,	 /* orange   */
	170,170,170, /* grey     */
	255,153,136, /* pink     */
	0,221,0,	 /* lt green */
	255,255,0,	 /* yellow   */
	0,255,153,	 /* aqua     */
	255,255,255};/* white    */

unsigned char ciderpresscolors[16][3] = {
	0,0,0,		 /* black    */
	221,0,51,	 /* red      */
	0,0,153,	 /* dk blue  */
	221,34,221,	 /* purple   */
	0,119,34,	 /* dk green */
	85,85,85,	 /* gray     */
	34,34,255,	 /* med blue */
	102,170,255, /* lt blue  */
	136,85,0,	 /* brown    */
	255,102,0,	 /* orange   */
	170,170,170, /* grey     */
	255,153,136, /* pink     */
	17,221,0,	 /* lt green */
	255,255,0,	 /* yellow   */
	68,255,153,	 /* aqua     */
	255,255,255};/* white    */

unsigned char awinoldcolors[16][3] = {
	0,0,0,       /* black    */
	208,0,48,	 /* red      */
	0,0,128,	 /* dk blue  */
	255,0,255,	 /* purple   */
	0,128,0,	 /* dk green */
	128,128,128, /* gray     */
	0,0,255,	 /* med blue */
	96,160,255,	 /* lt blue  */
	128,80,0,	 /* brown    */
    255,128,0,	 /* orange   */
    192,192,192, /* grey     */
    255,144,128, /* pink     */
    0,255,0,	 /* lt green */
    255,255,0,	 /* yellow   */
    64,255,144,	 /* aqua     */
    255,255,255};/* white    */

unsigned char awinnewcolors[16][3] = {
	0,0,0,       /* black    */
	157,9,102,   /* red      */
	42,42,229,   /* dk blue  */
	199,52,255,  /* purple   */
	0,118,26,    /* dk green */
	128,128,128, /* gray     */
	13,161,255,  /* med blue */
	170,170,255, /* lt blue  */
	85,85,0,     /* brown    */
    242,94,0,    /* orange   */
    192,192,192, /* grey     */
    255,137,229, /* pink     */
    56,203,0,    /* lt green */
    213,213,26,  /* yellow   */
    98,246,153,  /* aqua     */
    255,255,254};/* white    */

uchar shrpalette[32], rgbArray[16][3], rgbAppleArray[16][3];
double rgbLuma[16], rgbDouble[16][3];

/* ***************************************************************** */
/* ========================== code ================================= */
/* ***************************************************************** */

/* intialize the values for the current palette */
void InitDoubleArrays()
{
	int i;
	double dr, dg, db;
	unsigned r, g, b;

    /* array for matching closest sprite color in SHR palette */
	for (i=0;i<16;i++) {
		rgbDouble[i][0] = dr = (double) rgbArray[i][0];
		rgbDouble[i][1] = dg = (double) rgbArray[i][1];
		rgbDouble[i][2] = db = (double) rgbArray[i][2];
		rgbLuma[i] = (dr*299 + dg*587 + db*114) / (255.0*1000);
	}

}

/* use built-in palette if no shrfile is named */
sshort GetBuiltinPalette(sshort palidx)
{
	sshort i,j;
	uchar r,g,b;

	switch(palidx) {
		case 3: for (i=0;i<16;i++) {
					rgbArray[i][0] = awinnewcolors[i][0];
					rgbArray[i][1] = awinnewcolors[i][1];
					rgbArray[i][2] = awinnewcolors[i][2];
				}
				break;
		case 2: for (i=0;i<16;i++) {
					rgbArray[i][0] = awinoldcolors[i][0];
					rgbArray[i][1] = awinoldcolors[i][1];
					rgbArray[i][2] = awinoldcolors[i][2];
				}
				break;
		case 1: for (i=0;i<16;i++) {
					rgbArray[i][0] = ciderpresscolors[i][0];
					rgbArray[i][1] = ciderpresscolors[i][1];
					rgbArray[i][2] = ciderpresscolors[i][2];
				}
				break;
		default: for (i=0;i<16;i++) {
					rgbArray[i][0] = kegs32colors[i][0];
					rgbArray[i][1] = kegs32colors[i][1];
					rgbArray[i][2] = kegs32colors[i][2];
				}
				break;
		}


		for (i=0;i<16;i++) {
			rgbAppleArray[i][0] = r = rgbArray[i][0] >> 4;
			rgbAppleArray[i][1] = g = rgbArray[i][1] >> 4;
			rgbAppleArray[i][2] = b = rgbArray[i][2] >> 4;

			j = i*2;
			shrpalette[j+1] = r;
			shrpalette[j] = (g << 4) | b;

		}
}


/* read SHR palette 0 from $C1 $0000 SHR PIC File */
sshort GetSHRPalette(char *shrfile)
{
	FILE *fp;
	sshort i,j,status = INVALID;
	ulong filesize;
	uchar r,g,b;

    if((fp=fopen(shrfile,"rb"))==NULL) {
		printf("Error Opening %s!\n",shrfile);
		return status;
	}

	for (;;) {

		fseek(fp, 0L, SEEK_END);
		filesize = ftell(fp);

		if (filesize != (ulong)32768L) {
			printf("%s is not a raw SHR file!\n",shrfile);
			break;
		}
    	/* seek past image data to first palette in SHR file */
		fseek(fp,32256L,SEEK_SET);
		/* read SHR palette into buffer */
		if (fread((char *)&shrpalette[0],1,32,fp) != 32) {
			printf("Error Reading %s!\n",shrfile);
			break;
		}
		/* expand SHR palette from 12 bit to 24 bit gun values */
		/* and read into rgbArray */
		for (i=0;i<16;i++) {
			j = i*2;

			rgbAppleArray[i][0] = (shrpalette[j+1] &0xf);
			rgbAppleArray[i][1] = (shrpalette[j] &0xf0) >> 4;
			rgbAppleArray[i][2] = (shrpalette[j] &0xf);

			rgbArray[i][0] = (shrpalette[j+1] &0xf) << 4;
			rgbArray[i][1] = (shrpalette[j] &0xf0);
			rgbArray[i][2] = (shrpalette[j] &0xf) << 4;
		}
		status = SUCCESS;
		break;
	}
	fclose(fp);
	return status;

}


/* use CCIR 601 luminosity to get closest color in SHR palette */
uchar GetDrawColor(uchar r, uchar g, uchar b)
{
	uchar drawcolor, red, green, blue;
	double dr, dg, db, diffR, diffG, diffB, luma, lumadiff, distance, prevdistance;
	int i;

    red = (uchar)(r >> 4);
    green = (uchar)(g >> 4);
    blue = (uchar)(b >> 4);

    /* quick check for verbatim match */
	for (i = 0; i < 16; i++) {
		if (rgbAppleArray[i][0] == red &&
		    rgbAppleArray[i][1] == green &&
		    rgbAppleArray[i][2] == blue) return (uchar)i;

	}

    dr = (double)r;
    dg = (double)g;
    db = (double)b;
    luma = (dr*299 + dg*587 + db*114) / (255.0*1000);
    lumadiff = rgbLuma[0]-luma;

	/* Compare the difference of RGB values, weigh by CCIR 601 luminosity */
    /* set palette index to SHR color with shortest distance */

    /* get color distance to first SHR palette color */
	diffR = (rgbDouble[i][0]-dr)/255.0;
	diffG = (rgbDouble[i][1]-dg)/255.0;
	diffB = (rgbDouble[i][2]-db)/255.0;
    prevdistance = (diffR*diffR*0.299 + diffG*diffG*0.587 + diffB*diffB*0.114)*0.75
         + lumadiff*lumadiff;
    /* set palette index to first SHR color */
    drawcolor = 0;

    /* get color distance to rest of SHR palette colors */
    for (i=1;i<16;i++) {

        /* get color distance of to this SHR index */
		lumadiff = rgbLuma[i]-luma;
		diffR = (rgbDouble[i][0]-dr)/255.0;
		diffG = (rgbDouble[i][1]-dg)/255.0;
		diffB = (rgbDouble[i][2]-db)/255.0;
    	distance = (diffR*diffR*0.299 + diffG*diffG*0.587 + diffB*diffB*0.114)*0.75
         	+ lumadiff*lumadiff;

        /* if distance is smaller use this index */
	    if (distance < prevdistance) {
		   prevdistance = distance;
		   drawcolor = (uchar)i;
		}

	}
	return drawcolor;
}


/* reads a 24 bit BMP file in the range from 2 x 1 to 320 x 200 */
/* writes a $C1 $CC65 image fragment */
sshort Convert()
{

    FILE *fp,*fp2,*fp3;
    sshort status = INVALID, screenimage=INVALID;
	ushort x,y,i,spritewidth,spriteheight,packet;
	uchar r,g,b,c,c1,c2,bmpscanline[960];
	ulong pos;

    if((fp=fopen(bmpfile,"rb"))==NULL) {
		printf("Error Opening %s for reading!\n",bmpfile);
		return status;
	}
    /* read the header stuff into the appropriate structures */
    fread((char *)&bfi.bfType,
	             sizeof(BITMAPFILEHEADER),1,fp);
    fread((char *)&bmi.biSize,
                 sizeof(BITMAPINFOHEADER),1,fp);

    if (bmi.biCompression==BI_RGB &&
        bfi.bfType[0] == 'B' && bfi.bfType[1] == 'M' &&
        bmi.biPlanes==1 && bmi.biBitCount == 24) {

		spritewidth = (ushort) bmi.biWidth;
		spriteheight = (ushort) bmi.biHeight;

		if (spritewidth > 1 && spritewidth < 321 &&
			spriteheight > 0 && spriteheight < 201) status = SUCCESS;

	    if (spritewidth == 320 && spriteheight == 200) screenimage = SUCCESS;
	}

    if (status == INVALID) {
		fclose(fp);
		printf("%s is in the wrong format!\n",bmpfile);
		return status;
	}

    if((fp2=fopen(spritefile,"wb"))==NULL) {
		printf("Error Opening %s for writing!\n",spritefile);
		fclose(fp);
		return INVALID;
	}

	if (screenimage == SUCCESS) {
	   if((fp3=fopen(picfile,"wb"))==NULL) {
			printf("Error Opening %s for writing!\n",picfile);
			fclose(fp2);
			fclose(fp);
			return INVALID;
		}
	}

	packet = spritewidth * 3;
    /* BMP scanlines are padded to a multiple of 4 bytes (DWORD) */
	while ((packet % 4) != 0) packet++;
	/* SHR sprites are in SHR pixel-pairs - multiples of two
	   for fast plotting etc. */
	spritewidth = (spritewidth/2);

	/* write 2 byte header */
	if (writeheader == 1) {
		fputc((uchar)spritewidth,fp2);  /* width in bytes */
		fputc((uchar)spriteheight,fp2); /* height in pixels */
	}

    /* write header values to stdout */
	if (quietmode == 0) {
		printf("#define %sWIDTH  %d\n",fname,spritewidth);
		printf("#define %sHEIGHT %d\n",fname,spriteheight);
		printf("#define %sSIZE   %d\n\n",fname,spritewidth * spriteheight);

	}
    /* write 32 byte shr palette */
	if (writepalette == 1) {
		fwrite((char *)&shrpalette[0],1,32,fp2);
	}

    /* write shr palette array values to stdout */
	if (quietmode == 0) {
		printf("/* Embedded SHR Image Fragment created from %s */\n\n",bmpfile);
        printf("unsigned char %sColorTable[32] = {",fname);
		for (x = 0; x < 32; x++) {
			if (x == 0 || x == 16) {
				printf("\n%3d,",shrpalette[x]);
				continue;
			}
			if (x == 31) {
				printf("%3d};\n\n",shrpalette[x]);
				break;
			}
			printf("%3d,",shrpalette[x]);
		}
	}

	/* write background color */
	if (backgroundcolor > -1 && backgroundcolor < 16) {
		fputc((unsigned char)backgroundcolor,fp2);
		if (quietmode == 0) printf("unsigned char %sBackgroundColor = %d;\n\n",fname,backgroundcolor);

	}

	if (quietmode == 0) {
		printf("unsigned char %sPixelData[] = {\n",fname);
	}

	/* read BMP from top scanline to bottom scanline */
    pos = (ulong) (spriteheight - 1);
    pos *= packet;
    pos += bfi.bfOffBits;

	for (y=0;y<spriteheight;y++,pos-=packet) {
		fseek(fp,pos,SEEK_SET);
		fread((char *)&bmpscanline[0],1,packet,fp);
		for (x = 0,i = 0; x < spritewidth; x++) {

            /* make 12-bit color pixel-pairs from 24-bit RGB triples in BMP */
			b = bmpscanline[i]; i++;
			g = bmpscanline[i]; i++;
			r = bmpscanline[i]; i++;
			c1 = GetDrawColor(r,g,b); /* get nearest color */

			b = bmpscanline[i]; i++;
			g = bmpscanline[i]; i++;
			r = bmpscanline[i]; i++;
			c2 = GetDrawColor(r,g,b); /* get nearest color */

			c = (c1 << 4) | c2;

            /* write pixel-pair value to SHR image fragment */
			fputc(c,fp2);
			if (screenimage == SUCCESS) fputc(c,fp3);

            /* write pixel-pair value to stdout */
			if (quietmode == 0) {
			    if (y == 0 && x == 0) {
					printf("%3d",c);
				}
				else {
					printf(",");
					if (x%16 == 0) printf("\n");
					printf("%3d",c);
				}
			}

		}
	}

	if (quietmode == 0) {
		printf("};\n\n");
	}

	fclose(fp);
	fclose(fp2);
	if (screenimage == SUCCESS) {
		c = 0;
		for (y = 0;y < 256; y++) fputc(c,fp3);
		fwrite((char *)&shrpalette[0],1,32,fp3);
		for (y = 0;y < 480; y++) fputc(c,fp3);
		fclose(fp3);
		if (quietmode == 1)printf("%s created!\n",picfile);
	}

	return SUCCESS;

}


char *usage[] = {
	"Usage:   \"b2sprite sprite.bmp -options\" ",
	"*or*     \"b2sprite sprite.bmp background.shr -options\"",
	"Options: -t use CiderPress file attribute preservation tags.",
	"         Tags are turned-off by default.",
	"         -p (-p0,-p1,-p2,-p3) use built-in palette.",
	"         Kegs32 DHGR color palette (-p0) is the default.",
	"         background.shr over-rides built-in palette.",
	"         -b (-b0 to -b15) background color.",
	"         -np do not write palette to output.",
	"         -nh do not wrtite header to output.",
	"         -q disable quietmode (can be redirected to file).",
	"         creates embedded sprite array.",
	"Output:  sprite.sprite - $C1 $CC65 SHR image fragment.",
	NULL};

void pusage(void)
{
	sshort i;

	for (i=0;usage[i] != NULL;i++) puts(usage[i]);
}


int main(int argc, char **argv)
{
	sshort idx,jdx,palidx=0,tags=0;
	uchar ch, *wordptr;

    if (argc < 2) {
		pusage();
		return (1);
	}

    /* getopts */
    if (argc > 2) {
    	for (idx = 2; idx < argc; idx++) {
	   		wordptr = (uchar *)&argv[idx][0];
	   		ch = wordptr[0];
	   		if (ch != '-') {
				if (GetSHRPalette(wordptr) == SUCCESS) palidx = -1;
				continue;
			}
			ch = toupper(wordptr[1]);
		    switch(ch) {

				case 'B': jdx = atoi((char *)&wordptr[2]);
				          if (jdx > -1 && jdx < 16) backgroundcolor = jdx;
				          break;
				case 'P': if (palidx == -1) break;
				          jdx = atoi((char *)&wordptr[2]);
				          if (jdx > -1 && jdx < 4) palidx = jdx;
				          break;
				case 'Q': quietmode = 0;
				          break;
				case 'T': tags = 1;
				          break;
				case 'N': ch = toupper(wordptr[2]);
				          if (ch == 'P') writepalette = 0;
				          else if (ch == 'H') writeheader = 0;
				          break;

			}
		}
	}

    jdx = 999;
	strcpy(fname, argv[1]);
	for (idx = 0; fname[idx] != (uchar)0; idx++) {
		if (fname[idx] == '.') {
			jdx = idx;
		}
	}
    if (jdx != 999) fname[jdx] = (uchar)0;

    sprintf(bmpfile,"%s.bmp",fname);

    if (tags == 1) {
		sprintf(spritefile,"%s.SPRITE#C1CC65",fname);
		sprintf(picfile,"%s.SHR#C10000",fname);
	}
    else {
		sprintf(spritefile,"%s.SPRITE",fname);
		sprintf(picfile,"%s.SHR",fname);
	}

	if (palidx != -1)GetBuiltinPalette(palidx);
    InitDoubleArrays();
    if (Convert() == INVALID) return (1);

    if (quietmode == 1)printf("%s created!\n",spritefile);

	return SUCCESS;
}

